#! /usr/bin/env /usr/bin/python

import time
import imaplib
from ical import *
import re

import string
import random, StringIO, rfc822

class ConnectionErr(Exception):
    pass
class PrivilegeErr(Exception):
    pass
class ConfusionErr(Exception):
    pass



class schedule_calendar: 
    def __init__(self, server, user, passwd):
        self.parser = icalparser_new()
        if (None == self.parser):
            raise ConfusionErr, "parser object for ical format won't initialize"
        self.myserver = imaplib.IMAP4(server)
        if (None == self.myserver):
            raise ConnectionErr, "Cannot connect to server. Network/Server problem"
        self.myserver.debug = 0
        resp, items = self.myserver.login(user, passwd)
        if (None == re.search('^OK$', resp)):
            raise PriviledgeErr, "Cannot log on to server. Password/Account problem"
        self.loggedout=0
        self.internaluids=0
        self.messageflags={}
        self.uid={}
        self.internaldate={}
        self.msgtext={}
        self.parsed={}
        self.synced={}
        self.concordance={}
             
    def load(self):
#        resp, items = self.myserver.list(pattern= '*')
#        #items = string.split(items[0])
#        print "resp = {%s}" % resp
#        print "items = {%s}" % items
#        for i in items:
#            print "\titem = {%s}" % i

         resp, items = self.myserver.select(mailbox = "Calendar")
         if (None == re.search('^OK$', resp)):
             raise ConfusionErr, "Calendar mailbox doesn't exist on server"
         
         resp, items = self.myserver.search(None, "ALL")
         if (None == re.search('^OK$', resp)):
             raise ConfusionErr, "Server will not tell me what messages there are"
         if (0 == len(items[0][:])):
             list = ()
         else:
             list = string.split(items[0], " ")
         for i in list:
             print "i=%s" % i
             resp, entries = self.myserver.fetch(i, '(FLAGS UID INTERNALDATE)')
             if (None == re.search('^OK$', resp)):
                 raise ConfusionErr, "Server will not provide me with UIDs and Flags"
             entry = entries[0]
             #{1 (FLAGS (\Seen) UID 548 INTERNALDATE "13-Mar-2002 20:21:22 -0500" RFC822.SIZE 1409)}
             print entry
             flagtext = re.search("FLAGS (\([^\)]*\))", entry)
             if flagtext:
                 self.messageflags[i]=flagtext.group(1)
             else:
                 self.messageflags[i]="()"

             uidtext = re.search("UID +([^ ]*) ", entry)
             if uidtext:
                 self.uid[i]=uidtext.group(1)
             else:
                 #this is an error
                 self.uid[i]=""
                 
             internaldatetext = re.search("INTERNALDATE \"([^\"]*)\"", entry)
             if internaldatetext:
                 self.internaldate[i]=internaldatetext.group(1)
             else:
                 self.internaldate[i]=""

             print "flags=%s\t UID=%s\t internaldate=%s" % (self.messageflags[i], self.uid[i], self.internaldate[i])

             resp, entries = self.myserver.fetch(i, "(RFC822)")
             if (None == re.search('^OK$', resp)):
                 raise ConfusionErr, "Server will not send me message contents"
             #    resp, entries = self.myserver.fetch(i, "(FLAGS)")

             self.msgtext[i] = entries[0][1]
             self.parsed[i] = icalcomponent_new_from_string(self.msgtext[i])
             if (None == self.parsed[i]):
                 raise ConfusionErr, "Ical_parser cant make head nor tail of what the server called a message. Out of memory?"
             self.synced[i]=1
             
         self.make_uid_to_entry()
             
    def make_uid_to_entry(self):
        for i in self.uid.keys():
            self.concordance[self.uid[i]] = i

    def get_uids(self):
        return self.uid.values()
         
    def get_text(self, givenuid):
        return self.msgtext[self.concordance[givenuid]]

    def get_parsed(self, givenuid):
        return self.parsed[self.concordance[givenuid]]

    def get_flags(self, givenuid):
        return self.messageflags[self.concordance[givenuid]]

    def set_flags_withsync(self, givenuid, givenflags):
        self.messageflags[self.concordance[givenuid]] = givenflags
        resp, entries = self.myserver.store(self.concordance[givenuid], "FLAGS", '(\\Deleted)');
        if (None == re.search('^OK$', resp)):
            raise ConnectionErr, "Server refuses to change flags (Network down, another user expunged it out from under you ...)"
        
    def delete(self, givenuid):
        self.set_flags_withsync(givenuid, '(\\Deleted)')
        self.synced=1
        
    def expunge(self):
        resp, entries = self.myserver.expunge()
        if (None == re.search('^OK$', resp)):
            raise ConnectionErr, "Server refuses to expunge messages (Network/server down)"
 
    def set_dirty(self, givenuid, givenflags):
        self.synced[self.concordance[givenuid]] = 0
        
    def get_internaldate(self, givenuid):
        return self.internaldate[self.concordance[givenuid]]

    def new_entry(self, type):
        i = 'lllocalidentifier%d' % len(self.uid.keys())
        self.uid[i]="LocalUID %s" % self.internaluids
        self.internaluids += 1
        self.internaldate[i]=""
        self.messageflags[i]=""
        newentry = icalcomponent_new(ICAL_VCALENDAR_COMPONENT) 
        icalcomponent_set_method(newentry, ICAL_METHOD_REQUEST)
        tzone = icalcomponent_new_localtimezone()
        icalcomponent_add_component(newentry, tzone)
        newevent = icalcomponent_new(type)
        icalcomponent_add_component(newentry, newevent)
        self.parsed[i]=newentry
        self.synced[i]=0;
        self.concordance[self.uid[i]] = i
        return (self.uid[i], newevent)
    
    def sync_entry(self, givenuid):
        i = self.concordance[givenuid]
        s = icalcomponent_as_ical_string(self.parsed[i])
        subject = subject_in_parsed_message(self.parsed[i])
        date    = time.strftime('%a, %d %b %Y %H:%M:%S', time.localtime())
        if (time.localtime()[8]):
            tz = time.altzone/60
        else:
            tz = time.timezone/60
        if (0<tz):
            date += ' -'
        else:
            date += ' -'
            tz = 0-tz
        date += '%d' % ((tz/600) % 10)
        date += '%d' % ((tz/60) % 10)
        date += '%d' % ((tz/10) % 6)
        date += '%d' % (tz % 10)
        code    = '01C09B9D.0353D3D0'
        desc    = description_in_parsed_message(self.parsed[i])
        mime_header='MIME-Version: 1.0\nContent-Type: multipart/alternative;\n\tboundary="----_=_NextPart_001_%s"\ncontent-class: urn:content-classes:appointment\nSubject: %s\nDate: %s\n\nThis is a multi-part message in MIME format.\n\n------_=_NextPart_001_%s\nContent-Type: text/plain;\n\tcharset="iso-8859-1"\nContent-Transfer-Encoding: quoted-printable\n\n%s\n\n------_=_NextPart_001_%s\ncontent-class: urn:content-classes:appointment\nContent-Type: text/calendar;\n\tmethod=REQUEST;\n\tcharset="UTF-8"\nContent-Transfer-Encoding: 8bit\n\n' % (code, subject, date, code, desc, code)
        mime_tail='\n\n------_=_NextPart_001_%s--\n' % code

        unclean = re.compile('[\n] ', re.MULTILINE)
        (s, count) = unclean.subn('', s)
        print "count = %d" % count

        unclean = re.compile('\r', re.MULTILINE)
        (s, count) = unclean.subn('', s)
        print "count = %d" % count

        m = mime_header+s+mime_tail
        unclean = re.compile('\n', re.MULTILINE)
        (m, count) = unclean.subn('\r\n', m)

        print "=================================================================="
        print "%s" % m
        print "=================================================================="
        if (None == re.search('^LocalUID', givenuid)):
            self.delete(givenuid)
            self.uid[i]="LocalUID %s" % self.internaluids
            self.internaluids += 1
            self.concordance[self.uid[i]] = i
            
        resp, items = self.myserver.append(mailbox = "Calendar", message=m, date_time='"27-Mar-2002 20:21:22 -0500"', flags=self.messageflags[i])
        if (None == re.search('^OK$', resp)):
            raise ConnectionErr, "Server refuses to accept changes (Network down, another user expunged it out from under you ...)"

        #print "UID = {%s}" % string.split(items[0], " ")[2][0:-1]
        self.uid[i]=string.split(items[0], " ")[2][0:-1]
        self.synced=1
        return self.uid[i]
    
    def logout(self):
        resp, items = self.myserver.logout()
        if (None == re.search('^BYE$', resp)):
            raise ConnectionErr, "Server refused to acknowledge logout"
        self.loggedout = 1

    def free(self):
        if (0==self.loggedout):
            self.logout();
        self=None

def icalevent_add_uid(newevent, u=None):
    if (None == u):
        u = "0123456789ABCDEF0123456789ABCDEF0123456789ABCDEF0123456789ABCDEF"
    icalcomponent_add_property(newevent, icalproperty_new_uid(u))

def icalevent_add_organizer(newevent, email, commonname):
    whom = "MAILTO:" + email
    organizer = icalproperty_new_organizer(whom)
    icalproperty_add_parameter(organizer, icalparameter_new_cn(commonname))
    icalcomponent_add_property(newevent, organizer)

def icalevent_add_description(newevent, what):
    icalcomponent_add_property(newevent, icalproperty_new_description(what))

def icalevent_add_summary(newevent, what):
    icalcomponent_add_property(newevent, icalproperty_new_summary(what))

def icalevent_add_sequence(newevent, what=0):
    icalcomponent_add_property(newevent, icalproperty_new_sequence(what))

def icalproperty_add_localtimezoneid(c):
    icalproperty_add_parameter(c, icalparameter_new_tzid(time.tzname[0]))

def icalevent_add_dtstamp(newevent, when=0):
    if 0==when:
        when = time.time()
#    atime = icaltime_from_timetupple(time.localtime(when))
    atime = icaltime_from_timetupple(time.gmtime(when))
    dtstamp = icalproperty_new_dtstamp(atime)
    icalproperty_add_localtimezoneid(dtstamp)
    icalcomponent_add_property(newevent, dtstamp)

def icalevent_add_dtstart(newevent, when=0):
    if 0==when:
        when = time.time()
#    atime = icaltime_from_timetupple(time.localtime(when))
    atime = icaltime_from_timetupple(time.gmtime(when))
    dtstart = icalproperty_new_dtstart(atime)
    icalproperty_add_localtimezoneid(dtstart)
    icalcomponent_add_property(newevent, dtstart)

def icalevent_add_dtend(newevent, when=0):
    if 0==when:
        when = time.time()
#    atime = icaltime_from_timetupple(time.localtime(when))
    atime = icaltime_from_timetupple(time.gmtime(when))
    dtend = icalproperty_new_dtend(atime)
    icalproperty_add_localtimezoneid(dtend)
    icalcomponent_add_property(newevent, dtend)

def icalcomponent_new_localtimezone():
    tzone = icalcomponent_new(ICAL_VTIMEZONE_COMPONENT)
    icalcomponent_add_property(tzone, icalproperty_new_tzid(time.tzname[0]))

    tzonecomp = icalcomponent_new(ICAL_XSTANDARD_COMPONENT)
    icalcomponent_add_property(tzonecomp, icalproperty_new_dtstart(icaltime_from_string("16010101T020000")))
 
    icalcomponent_add_property(tzonecomp, icalproperty_new_tzoffsetfrom(-1 * time.altzone))
    icalcomponent_add_property(tzonecomp, icalproperty_new_tzoffsetto(-1 * time.timezone))
    rrabomination = icalproperty_new_rrule(icalrecurrencetype_from_string('FREQ=YEARLY;WKST=MO;INTERVAL=1;BYDAY=-1SU;BYMONTH=10'))
    icalcomponent_add_property(tzonecomp, rrabomination)
    icalcomponent_add_component(tzone, tzonecomp)

    if (0!=time.daylight):
        tzonecomp = icalcomponent_new(ICAL_XDAYLIGHT_COMPONENT)
        icalcomponent_add_property(tzonecomp, icalproperty_new_dtstart(icaltime_from_string("16010101T020000")))
        icalcomponent_add_property(tzonecomp, icalproperty_new_tzoffsetfrom(-1 * time.timezone))
        icalcomponent_add_property(tzonecomp, icalproperty_new_tzoffsetto(-1 * time.altzone))
        rrabomination = icalproperty_new_rrule(icalrecurrencetype_from_string('FREQ=YEARLY;WKST=MO;INTERVAL=1;BYDAY=-1SU;BYMONTH=10'))
        icalcomponent_add_property(tzonecomp, rrabomination)
        icalcomponent_add_component(tzone, tzonecomp)
    return tzone

def icaltimestring(when):
    return time.strftime("%Y%m%dT%H%M%S", when)

def isdst(when):
    return when[9]

def icaltime_from_timetupple(when):
    propnow = icalproperty_new_from_string(time.strftime("DTSTAMP:%Y%m%dT%H%M%SZ", when))
    return icalvalue_get_datetime(icalproperty_get_value(propnow))

def subject_in_parsed_message(mesg):
  c = icalcomponent_get_first_real_component(mesg)
  if (None == c):
      return "no real component. Wierd!"
  c = icalcomponent_get_first_property(c, ICAL_SUMMARY_PROPERTY)
  if (None == c):
      return "no subject property in mesg"
  s = icalproperty_get_summary(c)
  if (None == s):
      return "no subject in subject property. Wierd!"
  return s

def anything_in_parsed_message(mesg, what):   #anything is a misnomer, change it
  c = icalcomponent_get_first_real_component(mesg)
  if (None == c):
      return None
  c = icalcomponent_get_first_property(c, what)
  if (None == c):
      return None
  if (ICAL_ACTION_PROPERTY==what):
      result = icalproperty_get_action(c)
  if (ICAL_ATTACH_PROPERTY==what):
      result = icalproperty_get_attach(c)
  if (ICAL_ATTENDEE_PROPERTY==what):
      result = icalproperty_get_attendee(c)
  if (ICAL_CALSCALE_PROPERTY==what):
      result = icalproperty_get_calscale(c)
  if (ICAL_CATEGORIES_PROPERTY==what):
      result = icalproperty_get_categories(c)
  if (ICAL_CLASS_PROPERTY==what):
      result = icalproperty_get_class(c)
  if (ICAL_COMMENT_PROPERTY==what):
      result = icalproperty_get_comment(c)
  if (ICAL_COMPLETED_PROPERTY==what):
      result = icalproperty_get_completed(c)
  if (ICAL_CONTACT_PROPERTY==what):
      result = icalproperty_get_contact(c)
  if (ICAL_CREATED_PROPERTY==what):
      result = icalproperty_get_created(c)
  if (ICAL_DESCRIPTION_PROPERTY==what):
      result = icalproperty_get_description(c)
  if (ICAL_DTEND_PROPERTY==what):
      result = icalproperty_get_dtend(c)
  if (ICAL_DTSTAMP_PROPERTY==what):
      result = icalproperty_get_dtstamp(c)
  if (ICAL_DTSTART_PROPERTY==what):
      result = icalproperty_get_dtstart(c)
  if (ICAL_DUE_PROPERTY==what):
      result = icalproperty_get_due(c)
  if (ICAL_DURATION_PROPERTY==what):
      result = icalproperty_get_duration(c)
  if (ICAL_EXDATE_PROPERTY==what):
      result = icalproperty_get_exdate(c)
  if (ICAL_EXRULE_PROPERTY==what):
      result = icalproperty_get_exrule(c)
  if (ICAL_FREEBUSY_PROPERTY==what):
      result = icalproperty_get_freebusy(c)
  if (ICAL_GEO_PROPERTY==what):
      result = icalproperty_get_geo(c)
  if (ICAL_LASTMODIFIED_PROPERTY==what):
      result = icalproperty_get_lastmodified(c)
  if (ICAL_LOCATION_PROPERTY==what):
      result = icalproperty_get_location(c)
  if (ICAL_MAXRESULTS_PROPERTY==what):
      result = icalproperty_get_maxresults(c)
  if (ICAL_MAXRESULTSSIZE_PROPERTY==what):
      result = icalproperty_get_maxresultssize(c)
  if (ICAL_METHOD_PROPERTY==what):
      result = icalproperty_get_method(c)
  if (ICAL_ORGANIZER_PROPERTY==what):
      result = icalproperty_get_organizer(c)
  if (ICAL_PERCENTCOMPLETE_PROPERTY==what):
      result = icalproperty_get_percentcomplete(c)
  if (ICAL_PRIORITY_PROPERTY==what):
      result = icalproperty_get_priority(c)
  if (ICAL_PRODID_PROPERTY==what):
      result = icalproperty_get_prodid(c)
  if (ICAL_QUERY_PROPERTY==what):
      result = icalproperty_get_query(c)
  if (ICAL_QUERYNAME_PROPERTY==what):
      result = icalproperty_get_queryname(c)
  if (ICAL_RDATE_PROPERTY==what):
      result = icalproperty_get_rdate(c)
  if (ICAL_RECURRENCEID_PROPERTY==what):
      result = icalproperty_get_recurrenceid(c)
  if (ICAL_RELATEDTO_PROPERTY==what):
      result = icalproperty_get_relatedto(c)
  if (ICAL_REPEAT_PROPERTY==what):
      result = icalproperty_get_repeat(c)
  if (ICAL_REQUESTSTATUS_PROPERTY==what):
      result = icalproperty_get_requeststatus(c)
  if (ICAL_RESOURCES_PROPERTY==what):
      result = icalproperty_get_resources(c)
  if (ICAL_RRULE_PROPERTY==what):
      result = icalproperty_get_rrule(c)
  if (ICAL_SCOPE_PROPERTY==what):
      result = icalproperty_get_scope(c)
  if (ICAL_SEQUENCE_PROPERTY==what):
      result = icalproperty_get_sequence(c)
  if (ICAL_STATUS_PROPERTY==what):
      result = icalproperty_get_status(c)
  if (ICAL_SUMMARY_PROPERTY==what):
      result = icalproperty_get_summary(c)
  if (ICAL_TARGET_PROPERTY==what):
      result = icalproperty_get_target(c)
  if (ICAL_TRANSP_PROPERTY==what):
      result = icalproperty_get_transp(c)
  if (ICAL_TRIGGER_PROPERTY==what):
      result = icalproperty_get_trigger(c)
  if (ICAL_TZID_PROPERTY==what):
      result = icalproperty_get_tzid(c)
  if (ICAL_TZNAME_PROPERTY==what):
      result = icalproperty_get_tzname(c)
  if (ICAL_TZOFFSETFROM_PROPERTY==what):
      result = icalproperty_get_tzoffsetfrom(c)
  if (ICAL_TZOFFSETTO_PROPERTY==what):
      result = icalproperty_get_tzoffsetto(c)
  if (ICAL_TZURL_PROPERTY==what):
      result = icalproperty_get_tzurl(c)
  if (ICAL_UID_PROPERTY==what):
      result = icalproperty_get_uid(c)
  if (ICAL_URL_PROPERTY==what):
      result = icalproperty_get_url(c)
  if (ICAL_VERSION_PROPERTY==what):
      result = icalproperty_get_version(c)
  if (ICAL_X_PROPERTY==what):
      result = icalproperty_get_x(c)
  if (ICAL_XLICCLUSTERCOUNT_PROPERTY==what):
      result = icalproperty_get_xlicclustercount(c)
  if (ICAL_XLICERROR_PROPERTY==what):
      result = icalproperty_get_xlicerror(c)
  if (ICAL_XLICMIMECHARSET_PROPERTY==what):
      result = icalproperty_get_xlicmimecharset(c)
  if (ICAL_XLICMIMECID_PROPERTY==what):
      result = icalproperty_get_xlicmimecid(c)
  if (ICAL_XLICMIMECONTENTTYPE_PROPERTY==what):
      result = icalproperty_get_xlicmimecontenttype(c)
  if (ICAL_XLICMIMEENCODING_PROPERTY==what):
      result = icalproperty_get_xlicmimeencoding(c)
  if (ICAL_XLICMIMEFILENAME_PROPERTY==what):
      result = icalproperty_get_xlicmimefilename(c)
  if (ICAL_XLICMIMEOPTINFO_PROPERTY==what):
      result = icalproperty_get_xlicmimeoptinfo(c)
  
  return result

def description_in_parsed_message(mesg):
  c = icalcomponent_get_first_real_component(mesg)
  if (None == c):
      return "no real component. Wierd!"
  c = icalcomponent_get_first_property(c, ICAL_DESCRIPTION_PROPERTY)
  if (None == c):
      return "No description given"
  s = icalproperty_get_summary(c)
  if (None == s):
      return "no description in description property. Wierd!"
  return s
    
